/**
 * @author Olgierd Humeńczuk olgierdh@gmail.com
 * @file main
 */

#include <iostream>
#include <sstream>
#include <stdlib.h>
#include <stdexcept>
#include <vector>
#include <map>
#include <math.h>
#include <cassert>

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <SDL.h>
#else
#include <SDL/SDL.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>

#undef main

#include "math.hpp"
#include "mesh.hpp"
#include "renderer.hpp"
#include "scene.hpp"

/**
 * @brief Funkcja przetwarza zdarzenia pochodzace z kontrolerow takich jak:
 *          * klawiatura
 *          * mysz komputerowa
 *
 * @return true jeżeli aplikacja ma pozostać otwarta, false wpp
 */
bool processInput( void )
{
    // zmienna pomocnicza
    SDL_Event event;

    // czy są jakieś czekające na przetworzenie zdarzenia ?
    if ( SDL_PollEvent( &event ) )
    {
        // sprawdzenie typu zdarzenia
        switch( event.type )
        {
            // wyjście z aplikacji
            case SDL_QUIT:
                return false;

            // użytkownik nacisnął ktorys z klawiszy na klawiaturze
            case SDL_KEYDOWN:
                // jeżeli nacisnał klawisz eskape zamykamy aplikacje
                if( event.key.keysym.sym == SDLK_ESCAPE )
                {
                    return false;
                }
                break;
            case SDL_MOUSEMOTION:
                SDL_GetMouseState( &mousePosition.x, &mousePosition.y );
                break;
        }
    }
    return true;
}

/**
 * @brief Funkcja odpowiadająca za inicjalizację biblioteki SDL
 *          oraz utworzenie okna z kontekstem OpenGL.
 * @param scrWidth szerokosc ekranu
 * @param scrHeight wysokosc ekranu
 *
 * @throw std::runtime_error jeżeli nastąpi jakikolwiek błąd.
 */
void initSDL( int scrWidth, int scrHeight )
{
    // zaincjalizowanie części biblioteki odpowiadającej za wyświetlanie
    if( SDL_Init( SDL_INIT_VIDEO ) < 0 )
    {
        throw std::runtime_error( "Can't initialize SDL\n" );
    }

    // ustawienie ilości bitow dla każdego z kolorow
    SDL_GL_SetAttribute( SDL_GL_RED_SIZE, 8 );
    SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE, 8 );
    SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE, 8 );

    // ustawienie ilości bitow dla specjalnych buforow
    SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE, 24 );
    SDL_GL_SetAttribute( SDL_GL_STENCIL_SIZE, 8 );

    // wlaczenie podwojnego buforowania
    SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, true );
    SDL_GL_SetAttribute( SDL_GL_SWAP_CONTROL, 1 );

    // stworzenie okna aplikacji z kontekstem OpenGL'a
    if ( SDL_SetVideoMode( scrWidth, scrHeight, 24, SDL_OPENGL ) == 0 )
    {
        throw std::runtime_error(
            std::string( "Unable to set video mode: " ).append(
            	std::string( SDL_GetError() ) ) );
    }
}

/**
 * @brief Definicja wymiarow i materialow aktorow uzywanych w grze
 */
struct defObj
{
    /// deklaracja rozmiarow kuli
    static const GLfloat ballRadius;
    static const GLint ballSlices;
    static const GLint ballSegments;

    /// deklaracja rozmiarow plaszczyzny
    static const GLfloat planeWidth;
    static const GLfloat planeHeight;

    /// deklaracja rozmiarow sciany pn
    static const GLfloat wallNWidth;
    static const GLfloat wallNHeight;
    static const GLfloat wallNThickness;

    /// deklaracja rozmiarow sciany w/z
    static const GLfloat wallEWWidth;
    static const GLfloat wallEWHeight;
    static const GLfloat wallEWThickness;

    /// deklaracja rozmiarow rakietki
    static const GLfloat paddleWidth;
    static const GLfloat paddleHeight;
    static const GLfloat paddleThickness;

    /// deklaracja rozmiarow cegly
    static const GLfloat brickWidth;
    static const GLfloat brickHeight;
    static const GLfloat brickThickness;

    /// deklaracja materialow aktorow gry
    static const material planeMaterial;
    static const material wallNMaterial;
    static const material wallEWMaterial;
    static const material ballMaterial;
    static const material paddleMaterial;
    static const material brickMaterial;
};

/// definicja rozmiarow kuli
const GLfloat defObj::ballRadius = 1.4f;
const GLint defObj::ballSlices = 20;
const GLint defObj::ballSegments = 20;

/// definicja rozmiarow plaszczyzny
const GLfloat defObj::planeWidth = 100.0f;
const GLfloat defObj::planeHeight = 80.0f;

/// definicja rozmiarow sciany pn
const GLfloat defObj::wallNWidth = 100.0f;
const GLfloat defObj::wallNHeight = 4.0f;
const GLfloat defObj::wallNThickness = 4.0f;

/// definicja rozmiarow sciany w/z
const GLfloat defObj::wallEWWidth = 4.0f;
const GLfloat defObj::wallEWHeight = 4.0f;
const GLfloat defObj::wallEWThickness = 80.0f;

/// definicja rozmiarow rakietki
const GLfloat defObj::paddleWidth = 8.0f;
const GLfloat defObj::paddleHeight = 2.0f;
const GLfloat defObj::paddleThickness = 2.0f;

/// definicja rozmiarow cegly
const GLfloat defObj::brickWidth = 8.0f;
const GLfloat defObj::brickHeight = 3.0f;
const GLfloat defObj::brickThickness = 3.0f;

const material defObj::planeMaterial = material(
          vector4f( .1f, .1f, .1f, 1.f )
        , vector4f( .8f, .8f, .8f, 1.f )
        , vector4f( .8f, .8f, .8f, 1.f )
        , 20.0f );

const material defObj::wallNMaterial = material(
          vector4f( .1f, .1f, .1f, 1.f )
        , vector4f( .8f, .8f, .8f, 1.f )
        , vector4f( .8f, .8f, .8f, 1.f )
        , 20.0f );

const material defObj::wallEWMaterial = material(
          vector4f( .1f, .1f, .1f, 1.f )
        , vector4f( .8f, .8f, .8f, 1.f )
        , vector4f( .8f, .8f, .8f, 1.f )
        , 20.0f );

const material defObj::ballMaterial = material(
          vector4f( .2f, .0f, .0f, 1.f )
        , vector4f( .8f, .0f, .0f, 1.f )
        , vector4f( .8f, .0f, .0f, 1.f )
        , 70.0f );

const material defObj::paddleMaterial = material(
          vector4f( .0f, .0f, .6f, 1.f )
        , vector4f( .0f, .0f, .6f, 1.f )
        , vector4f( .0f, .0f, .6f, 1.f )
        , 10.0f );

const material defObj::brickMaterial = material(
          vector4f( .6f, .6f, .2f, 1.f )
        , vector4f( .6f, .6f, .2f, 1.f )
        , vector4f( .6f, .6f, .2f, 1.f )
        , 50.0f );

/// definicja typu zarzadcy sceny w naszej grze
typedef sceneManager< defObj > arkanoidSceneManager;

/**
 * @struct sceneUpdater
 * @brief Funktor odpowiedzialny za aktualizacje sceny
 */
struct sceneUpdater
{
    /**
     * @brief Przeciazony operator funkcyjny
     * @param object obiekt ktorego pozycja ma zostac zaktualizowana
     */
    void operator()( arkanoidSceneManager::sceneObjectsContainer::value_type& object )
    {
        baseSceneObject& obj = object.second;
        switch( obj.objectType )
        {
            case OBJECT_TYPE_BALL:
                obj.position += obj.dir * obj.velocity;
                break;
            case OBJECT_TYPE_PADDLE:
                obj.position.x = transformPos( static_cast< GLfloat >( mousePosition.x ),
                    static_cast< GLfloat >( 800 )
                  , -1 * ( defObj::planeWidth ) * 0.5f + defObj::paddleWidth * .5f
                  , ( defObj::planeWidth ) * 0.5f - defObj::paddleWidth * .5f );
                break;
        }
    }
};

/**
 * @brief Głowna metoda programu ( entry point )
 * @param argc ilosc argumentow przekazanych do programu
 * @param argv tablica wartosci argumentow
 */
int main( const int /*argc*/, const char** /*argv*/ )
{
    const int screenX = 800;
	const int screenY = 600;

    try
    {
        initSDL( screenX, screenY );
        initOpenGL( screenX, screenY );

        arkanoidSceneManager my_manager;

        my_manager.addSceneObject( OBJECT_TYPE_PLANE, vector3f( .0f, .0f, .0f ) );

        int32_t ballId = my_manager.addSceneObject(
                  OBJECT_TYPE_BALL, vector3f( 0.0f, defObj::ballRadius, 30.0f )
                , vector3f( 1.0f, 0.0f, -1.0f ).normalize(), 0.5f );

        int32_t paddleId = my_manager.addSceneObject(
                OBJECT_TYPE_PADDLE, vector3f( .0f, defObj::paddleHeight * 0.5f, 39.0f ) );
        int32_t wallNId = my_manager.addSceneObject(
                OBJECT_TYPE_WALL_N, vector3f( .0f, defObj::wallNHeight * 0.5f, -42.0f ) );
        int32_t wallEId = my_manager.addSceneObject(
                OBJECT_TYPE_WALL_EW, vector3f( -52.0f, defObj::wallEWHeight * 0.5f, 0.0f ) );
        int32_t wallWId = my_manager.addSceneObject(
                OBJECT_TYPE_WALL_EW, vector3f( 52.0f, defObj::wallEWHeight * 0.5f, 0.0f ) );
        int32_t brickId = my_manager.addSceneObject(
                OBJECT_TYPE_BRICK, vector3f( 0.0f, defObj::brickHeight * 0.5f, -10.0f ) );


        glEnableClientState( GL_VERTEX_ARRAY );
        glEnableClientState( GL_NORMAL_ARRAY );

        // głowna pętla
        while( processInput() )
        {
            my_manager.updateScene< sceneUpdater >();

            glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
            glMatrixMode( GL_MODELVIEW );
            glLoadIdentity();

            glPushMatrix();

            gluLookAt( 0.0f, 90.0f, 90.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.5f, -0.5f );
            {
                //dane światła
                GLfloat position[] = { 0.0f, 70.0f, 0.0f, 1.0f };
                GLfloat ambient[] = { 0.8f, 0.8f, 0.8f, 1.0f };
                GLfloat diffuse[] = { 0.8f, 0.8f, 0.8f, 1.0f };
                GLfloat specular[] = { 0.8f, 0.8f, 0.8f, 1.0f };

                setPointLight( GL_LIGHT0, position, ambient, diffuse, specular );
            }

            my_manager.renderAllObjects();

            glPopMatrix();
            SDL_GL_SwapBuffers();
        }
    }
    catch( const std::exception& e )
    {
        std::cerr << "Exception caught: " << e.what() << std::endl;
    }

    SDL_Quit();
    return 0;
}
