#ifndef __MATH_HPP__
#define __MATH_HPP__

// globalna wartość PI
const double PI = 3.141592;
const double T_PI = PI * 2.0;
const double EPSILON = 0.001;

/**
 * @struct position2d
 * @brief pozycja dwuwymiarowa
 */
struct position2d
{
    /**
     * @brief domyslny konstr.
     * inicjalizuje wsp. wartosciami zerowymi
     */
    position2d( void ) : x( 0 ), y( 0 ) { }

    /// wspolrzedna x
    int x;

    /// wspolrzedna y
    int y;

} mousePosition;



/**
 * @struct vector3f
 * @brief Struktura przechowująca współrzędne wektora
 */
struct vector3f
{

    /**
     * @brief Domyslny konstruktor
     * Ustawia wszystkie wspolrzedne na ( 0, 0, 0 )
     */
    vector3f( void ) : x( .0f ), y( .0f ), z( .0f )
    {
    }

    /**
     * @brief Konstruktor ustawiajacy wspolrzedne
     * @param in_x wspolrzedna x
     * @param in_y wspolrzedna y
     * @param in_z wspolrzedna z
     */
    vector3f( GLfloat in_x, GLfloat in_y, GLfloat in_z )
        : x( in_x ), y( in_y ), z( in_z )
    {
    }

    /**
     * @brief normalizacja wektora
     * @return znormalizowany wektor
     */
    vector3f normalize( void ) const
    {
        GLfloat len = length();
        assert( len > 0 );

        GLfloat one_div_len = 1.0f / len;

        return vector3f(
            x * one_div_len,
            y * one_div_len,
            z * one_div_len );
    }

    /**
     * @brief obliczanie iloczynu skalarnego
     */
    float dot( const vector3f& in ) const
    {
        return x * in.x + y * in.y + z * in.z;
    }

    /**
     * @brief operator -
     */
    vector3f operator-( const vector3f& in ) const
    {
        return vector3f( x - in.x, y - in.y, z - in.z );
    }

    /**
     * @brief operator +
     */
    vector3f operator+( const vector3f& in ) const
    {
        return vector3f( x + in.x, y + in.y, z + in.z );
    }

    /**
     * @brief operator -
     */
    vector3f operator-( float scalar ) const
    {
        return vector3f( x - scalar, y - scalar, z - scalar );
    }

    /**
     * @brief operator +
     */
    vector3f operator+( float scalar ) const
    {
        return vector3f( x + scalar, y + scalar, z + scalar );
    }

    /**
     * @brief operator *
     */
    vector3f operator*( float scalar ) const
    {
        return vector3f( x * scalar, y * scalar, z * scalar );
    }

    /**
     * @brief operator /
     */
    vector3f operator/( float scalar ) const
    {
        return vector3f( x / scalar, y / scalar, z / scalar );
    }

    /**
     * @brief operator -
     */
    vector3f operator-( void ) const
    {
        return vector3f( -x, -y, -z );
    }

    /**
     * @brief operator -=
     */
    vector3f& operator-=( const vector3f& in )
    {
        x -= in.x;
        y -= in.y;
        z -= in.z;
        return *this;
    }

    /**
     * @brief operator +=
     */
    vector3f& operator+=( const vector3f& in )
    {
        x += in.x;
        y += in.y;
        z += in.z;
        return *this;
    }

    /**
     * @brief liczenie dlugosci wektora
     * @return dlugosc wektora
     */
    GLfloat length( void ) const
    {
        return sqrtf( x * x + y * y + z * z );
    }

    /**
     * @brief odbicie wektora względem podanego wektora normalnego
     */
    vector3f reflect( const vector3f normal )
    {
        vector3f ret( *this );
        ret = *this - normal * 2 * ( normal.dot( *this ) );
        return ret;
    }

    ///wspolrzedna x
    GLfloat x;
    ///wspolrzedna y
    GLfloat y;
    ///wspolrzedna z
    GLfloat z;
};


/**
 * @struct vector4f
 * @brief Struktura przechowująca współrzędne wektora ( czterowymiarowego )
 */
struct vector4f
{

    /**
     * @brief Domyslny konstruktor
     * Ustawia wszystkie wspolrzedne na ( 0, 0, 0, 0 )
     */
    vector4f( void ) : x( .0f ), y( .0f ), z( .0f ), w( .0f )
    {
    }

    /**
     * @brief Konstruktor ustawiajacy wspolrzedne
     * @param in_x wspolrzedna x
     * @param in_y wspolrzedna y
     * @param in_z wspolrzedna z
     * @param in_w wspolrzedna w
     */
    vector4f( GLfloat in_x, GLfloat in_y, GLfloat in_z, GLfloat in_w )
        : x( in_x ), y( in_y ), z( in_z ), w( in_w )
    {
    }

    ///wspolrzedna x
    GLfloat x;
    ///wspolrzedna y
    GLfloat y;
    ///wspolrzedna z
    GLfloat z;
    ///wspolrzedna w
    GLfloat w;
};

/**
 * @brief Funkcja pomocnicza, przelicza pozycję z jednej przestrzeni na inna
 *        okreslona podanymi parametrami
 *
 * @param posX pozycja wejściowa
 * @param maxPosX maksymalna wartosc jaką może uzyskać parametr posX
 * @param dstPosMin minimalna wartosc pozycji w docelowej przestrzeni
 * @param dstPosMax maksymalna wartosc pozycji w docelowej przestrzeni
 */
GLfloat transformPos( GLfloat posX, GLfloat maxPosX,
							 GLfloat dstPosMin, GLfloat dstPosMax )
{
    GLfloat normalized = posX / maxPosX;
    return dstPosMin + ( ( dstPosMax - dstPosMin ) * normalized );
}

/**
 * @brief Testowanie przecięcia okręgu z odcinkiem
 * @param sphereCenter srodek okregu
 * @param sphereRadius promien sfery
 * @param segBeg poczatek odcinka
 * @param segEnd koniec odcinka
 * @param closest najblizszy punkt na prostej przechodzacej przez poczatek i koniec
 *          testowanego odcinka
 * @param offset przesuniecie kuli wzgledem "prawdziwej" kolizji
 */
bool testSphereSegInter(
      const vector3f& sphereCenter, const float sphereRadius
    , const vector3f& segBeg, const vector3f& segEnd
    , vector3f& closest, vector3f& offset )

{
    vector3f seg = segEnd - segBeg;
    vector3f pt = sphereCenter - segBeg;

    float projLen = pt.dot( seg.normalize() );

    vector3f proj = seg.normalize() * projLen;
    closest = segBeg + proj;

    if( projLen <= 0 ) closest = segBeg;
    if( projLen >= seg.length() ) closest = segEnd;

    vector3f dist = ( sphereCenter - closest );

    float distLen = dist.length();
    offset = dist.normalize() * ( sphereRadius - distLen );

    return distLen <= sphereRadius;
}



#endif // __MATH_HPP__
